package hide.comp.cdb;
using hide.tools.Extensions;

import cdb.Data;

class ModalColumnForm extends Modal {

	var contentModal : Element;
	var form : Element;
	var editor : Editor;
	var sheet : cdb.Sheet;

	public function new( editor : Editor, sheet : cdb.Sheet, column : cdb.Data.Column, ?parent,?el) {
		super(parent,el);

		var editForm = (column != null);
		var base = editor.base;
		this.editor = editor;
		this.sheet = sheet;

		contentModal = new Element("<div tabindex='0'>").addClass("content-modal").appendTo(content);

		if (editForm)
			new Element("<h2> Edit column </h2>").appendTo(contentModal);
		else
			new Element("<h2> Create column </h2>").appendTo(contentModal);
		new Element("<p id='errorModal'></p>").appendTo(contentModal);

		form = new Element('<form id="col_form" onsubmit="return false">

			<table>
				<tr>
				<td class="first">Column name
				<td><input type="text" name="name" required/>
				</tr>

				<tr>
				<td>Column type
				<td>
				<select name="type">
				<option value="">---- Choose -----</option>
				<option value="id">Unique Identifier</option>
				<option value="string">Text</option>
				<option value="bool">Boolean</option>
				<option value="int">Integer</option>
				<option value="float">Float</option>
				<option value="enum">Enumeration</option>
				<option value="flags">Flags</option>
				<option value="ref">Reference</option>
				<option value="list">List</option>
				<option value="properties">Properties</option>
				<option value="color">Color</option>
				<option value="file">File</option>
				<option value="image">Image</option>
				<option value="tilepos">Tile</option>
				<option value="dynamic">Dynamic</option>
				<option value="layer">Data Layer</option>
				<option value="tilelayer">Tile Layer</option>
				<option value="gradient">Gradient</option>
				<option value="curve">Curve</option>
				<option value="custom">Custom Type</option>
				<option value="guid">GUID</option>
				</select>
				</tr>

				<tr class="values">
				<td>Possible Values
				<td><div contenteditable="true" id="values" class="custom-text-edit-standalone" style="min-height: 3em; max-width: 170px;"></div>
				</tr>

				<tr class="sheet">
				<td>Sheet
				<td><select name="sheet"></select>
				</tr>

				<tr class="disp">
				<td>Display
				<td>
					<select name="display">
					<option value="0">Default</option>
					<option value="1">Percentage</option>
					</select>
				</tr>

				<tr>
					<td>Default Value
					<td id=defaultValue>
				</tr>

				<tr class="kind">
					<td>Kind
					<td>
					<select name="kind">
					<option value="">Default</option>
					<option value="localizable">Localizable</option>
					<option value="script">Script</option>
					</select>
				</tr>

				<tr class="custom">
					<td>Type
					<td><select name="ctype"></select>
				</tr>

				<tr class="scope">
					<td>Scope
					<td>
					<select name="scope">
					<option value="">Global</option>
					</select>
				</tr>

				<tr class="formula hide">
					<td>Formula</td>
					<td>
						<select name="formula">
						<option value="">None</option>
						</select>
						<label><input type="checkbox" name="export" style="float:none;display:inline-block" checked/>&nbsp;Export</label>
					</td>
				</tr>

				<tr class="doc hide">
					<td>&nbsp;
					<td><label><input type="checkbox" name="hidden"/>&nbsp;Hidden</label>
				</tr>

				<tr class="doc hide">
					<td>&nbsp;
					<td><label><input type="checkbox" name="copyPasteImmutable"/>&nbsp;Ignore copy paste</label>
				</tr>

				<tr class="doc hide">
					<td>Documentation
					<td><textarea name="doc" rows=10 columns=70></textarea>
				</tr>

				<tr class="more">
					<td>
						<a href="#" class="doctog can-hide">[+]</a>
						<a href="#" class="doctog hide">[-]</a>
					</td>
				</tr>

				<tr class="opt">
					<td>&nbsp;
					<td><label><input type="checkbox" name="req"/>&nbsp;Required</label>
				</tr>

				<tr>
					<td>&nbsp;
					<td>
						<p class="buttons">
							<input class="edit" type="submit" value="Modify" id="editBtn" />
							<input class="create" type="submit" value="Create" id="createBtn" />
							<input type="submit" value="Cancel" id="cancelBtn" />
						</p>
				</tr>
			</table>

			</form>').appendTo(contentModal);

		var parent = sheet.getParent();
		if( parent == null )
			form.find(".scope").remove();
		else {
			var scope = 1;
			var scopes = form.find("[name=scope]");
			var p = parent;
			while( p != null ) {
				if( p.s.idCol != null )
					new Element("<option>").attr("value",""+scope).text(p.s.name).appendTo(scopes);
				p = p.s.getParent();
				scope++;
			}
		}

		var sheets = form.find("[name=sheet]");
		sheets.empty();
		var options = [];
		for( i in 0...base.sheets.length ) {
			var s = base.sheets[i];
			if( s.idCol == null ) continue;
			if( s.idCol.scope != null ) {
				var checkSheets = [sheet];
				for( c in sheet.columns )
					switch( c.type ) {
					case TRef(s): checkSheets.push(base.getSheet(s));
					default:
					}
				var found = false;
				for( sscope in checkSheets )
					if( Cell.isIdScope(s, sscope) ) {
						found = true;
						break;
					}
				if( !found )
					continue;
			}
			options.push(new Element("<option>").attr("value", "" + i).text(s.name)); // .appendTo(sheets);
		}
		options.sort((e1, e2) -> e1.text() > e2.text() ? 1 : -1);
		for (e in options) {
			e.appendTo(sheets);
		}

		var types = form.find("[name=type]");
		var req = form.find("[name=req]");
		var defaultValueEl = form.find("#defaultValue");
		function updateDefaultValueField() {
			defaultValueEl.empty();
			var hide = !req.is(':checked');
			if (req.is(':checked')) {
				switch (types.val()) {
					case "float", "int":
						new Element('<input name="defaultValue" type="number"/>').appendTo(defaultValueEl);
					case "bool":
						new Element('<input name="defaultValue" type="checkbox"/>').appendTo(defaultValueEl);
					case "string":
						new Element('<input name="defaultValue" />').appendTo(defaultValueEl);
					default:
						hide = true;
				}
			}

			defaultValueEl.parent().css({display: hide ? 'none' : 'table-row'});
		}

		function changeFieldType() {
			form.find("table").attr("class","").toggleClass("t_"+types.val());
			updateDefaultValueField();
		}
		types.change(function(_) changeFieldType());
		req.change(function(_) updateDefaultValueField());
		changeFieldType();

		var ctypes = form.find("[name=ctype]");
		new Element("<option>").attr("value", "").text("--- Select ---").appendTo(ctypes);
		for( t in base.getCustomTypes() )
			new Element("<option>").attr("value", "" + t.name).text(t.name).appendTo(ctypes);

		var cforms = form.find("[name=formula]");
		for( f in editor.formulas.getList(sheet) )
			new Element("<option>").attr("value", f.name).text(f.name).appendTo(cforms);

		function toggleHide() {
			form.find(".can-hide").toggleClass("hide");
		}
		form.find(".hide").addClass("can-hide");
		form.find(".doctog").click(function(_) toggleHide());
		form.find(".doctog").keydown(function(e) {
			if( e.keyCode == hxd.Key.ENTER ) {
				e.stopPropagation();
				form.find(".doctog").click();
			}
		});

		if (editForm) {
			form.addClass("edit");
			form.find("[name=name]").val(column.name);
			form.find("[name=type]").val(column.type.getName().substr(1).toLowerCase()).change();
			form.find("[name=req]").prop("checked", !column.opt);
			form.find("[name=display]").val(column.display == null ? "0" : Std.string(column.display));
			form.find("[name=kind]").val(column.kind == null ? "" : ""+column.kind);
			form.find("[name=scope]").val(column.scope == null ? "" : ""+column.scope);
			form.find("[name=hidden]").prop("checked", column.kind == Hidden);
			var p = Editor.getColumnProps(column);
			form.find("[name=copyPasteImmutable]").prop( "checked", p.copyPasteImmutable );
			if( column.documentation != null ) {
				form.find("[name=doc]").val(column.documentation);
				form.find(".doc").removeClass("hide");
			}
			updateDefaultValueField();
			switch( column.type ) {
			case TEnum(values), TFlags(values):
				form.find("#values")[0].innerText = values.join("\n");
			case TRef(sname), TLayer(sname):
				var index = base.sheets.indexOf(base.getSheet(sname));
				form.find("[name=sheet]").val( "" + index);
			case TCustom(name):
				form.find("[name=ctype]").val(name);
			case TInt, TFloat:
				form.find("[name=formula]").val( p.formula == null ? "" : p.formula );
				form.find("[name=export]").prop( "checked", !p.ignoreExport );
				form.find("[name=defaultValue]").val( column.defaultValue );
			case TBool:
				form.find("[name=defaultValue]").prop('checked', column.defaultValue );
			default:
				form.find("[name=defaultValue]").val( column.defaultValue );
			}
		} else {
			form.addClass("create");
			form.find("input").not("[type=submit]").val("");
			var isProp = sheet.parent != null && sheet.parent.sheet.columns[sheet.parent.column].type == TProperties;
			form.find("[name=req]").prop("checked", !isProp);
			form.find("[name=kind]").val("");
		}

		form.find("[name=name]").focus();

		contentModal.keydown(function(e) {
			if( e.keyCode == 27 )
				closeModal();
			if( e.keyCode == hxd.Key.ENTER && !js.Browser.document.activeElement.hasAttribute("contenteditable")) {
				if( e.target != form.find("[name=doc]")[0] ) {
					if (editForm)
						form.find("#editBtn").click();
					else
						form.find("#createBtn").click();
				}
			}
			e.stopPropagation();
		});
		contentModal.keypress(function(e) e.stopPropagation());
		contentModal.click( function(e) e.stopPropagation());

		form.find("#cancelBtn").click(function(e) closeModal());
		if( column != null && Editor.getColumnProps(column).formula != null )
			toggleHide();
	}

	public function setCallback(callback : (Void -> Void)) {
		form.find("#createBtn").click(function(e) {
			e.preventDefault();
			callback();
		});
		form.find("#editBtn").click(function(e) {
			e.preventDefault();
			callback();
		});
	}

	public function closeModal() {
		content.empty();
		close();
	}

	var validIdentifierReg = ~/^[_a-zA-Z][_a-zA-Z0-9]*$/;

	public function getColumn( refColumn : cdb.Data.Column) : Column{
		var base = editor.base;
		var v : Dynamic<String> = { };
		var cols = form.find("input, select, textarea").not("[type=submit]");
		for( i in cols.elements() )
			Reflect.setField(v, i.attr("name"), i.attr("type") == "checkbox" ? (i.is(":checked")?"on":null) : i.val());
		var divs = form.find("div[contenteditable]");
		for (i in divs.elements()) {
			Reflect.setField(v, i[0].id, i[0].innerText);
		}

		function stringToValues(str: String) {
			var vals = StringTools.trim(str).split("\n");
			vals = [for ( v in vals) for (e in v.split(",")) e];
			vals = [for (v in vals) StringTools.trim(v)];
			for (v in vals) {
				if (!validIdentifierReg.match(v)) {
					error('Invalid name : "$v"');
					return null;
				}
			}
			if( vals.length == 0 ) {
				error("Missing value list");
				return null;
			}
			return vals;
		}

		var t : ColumnType = switch( v.type ) {
		case "id": TId;
		case "int": TInt;
		case "float": TFloat;
		case "string": TString;
		case "bool": TBool;
		case "enum":
			var vals = stringToValues(v.values);
			if (vals == null)
				return null;
			TEnum([for( f in vals ) StringTools.trim(f)]);
		case "flags":
			var vals = stringToValues(v.values);
			if (vals == null)
				return null;
			if( vals.length > 30 ) {
				error("Too many possible values");
				return null;
			}
			TFlags([for( f in vals ) StringTools.trim(f)]);
		case "ref":
			var s = base.sheets[Std.parseInt(v.sheet)];
			if( s == null ) {
				error("Sheet not found");
				return null;
			}
			TRef(s.name);
		case "image":
			TImage;
		case "list":
			TList;
		case "custom":
			var t = base.getCustomType(v.ctype);
			if( t == null ) {
				error("Type not found");
				return null;
			}
			TCustom(t.name);
		case "color":
			TColor;
		case "layer":
			var s = base.sheets[Std.parseInt(v.sheet)];
			if( s == null ) {
				error("Sheet not found");
				return null;
			}
			TLayer(s.name);
		case "file":
			TFile;
		case "tilepos":
			TTilePos;
		case "tilelayer":
			TTileLayer;
		case "gradient":
			TGradient;
		case "curve":
			TCurve;
		case "dynamic":
			TDynamic;
		case "properties":
			TProperties;
		case "guid":
			TGuid;
		default:
			return null;
		}
		var c : Column = {
			type : t,
			typeStr : null,
			name : v.name,
		};
		c.opt = v.req != "on";
		if( v.display != "0" ) c.display = cast Std.parseInt(v.display);
		c.kind = js.Lib.undefined;
		switch( v.kind ) {
		case "localizable": c.kind = Localizable;
		case "script": c.kind = Script;
		}
		if( form.find("[name=hidden]").is(":checked") ) c.kind = Hidden;

		var props = Editor.getColumnProps(c);
		props.copyPasteImmutable = form.find("[name=copyPasteImmutable]").is(":checked") ? true : null;
		switch( t ) {
		case TFloat, TInt:
			props.formula = form.find("[name=formula]").val();
			if( props.formula == "" ) props.formula = null;
			props.ignoreExport = props.formula != null && !form.find("[name=export]").is(":checked") ? true : null;
			var defV = form.find("[name=defaultValue]").val();
			if (defV != null && defV != '')
				c.defaultValue = Std.parseFloat(defV);
			else
				Reflect.deleteField(c, "defaultValue");
		case TBool:
			var defV = form.find("[name=defaultValue]").is(':checked');
			if (defV != null && defV)
				c.defaultValue = true;
			else
				Reflect.deleteField(c, "defaultValue");
		default:
			var defV = form.find("[name=defaultValue]").val();
			if (defV != null && defV != '')
				c.defaultValue = defV;
			else
				Reflect.deleteField(c, "defaultValue");
		}
		if( t == TId && v.scope != "" ) c.scope = Std.parseInt(v.scope);
		if( v.doc != "" ) c.documentation = v.doc;

		var hasProp = false;
		for( f in Reflect.fields(props) )
			if( Reflect.field(props,f) == null )
				Reflect.deleteField(props, f);
			else
				hasProp = true;
		c.editor = hasProp ? props : js.Lib.undefined;
		return c;
	}

	public function error(str : String) {
		contentModal.find("#errorModal").html(str);
	}
}